
#ifndef __events_hpp
#define __events_hpp

#include "core.h"

/**
 * High priority listener (value = 100)
 */
#define HIGH_PRIORITY 100

/**
 * Medium priority listener (value = 50)
 */
#define MEDIUM_PRIORITY 50

/**
 * Normal priority listener (value = 0)
 */
#define NORMAL_PRIORITY 0

/**
 * Low priority listener (value = -100)
 */
#define LOW_PRIORITY -100

namespace capi
{
	// Pre-declaration of classes
	template <typename T> class EventType;
	template <typename T> class EventLinker;
	template <typename T> class AbstractEvent;
	class EventDispatcher;
	class Event;
	class EventListener;

    using namespace std;

	/**
	 * \brief Type of an event (used to attached listeners)
	 *
	 * Represents an event type. An EventType is unique and can
	 * be associated with many events and listeners.
	 *
	 * \author	Cedric Tabin
	 * \date	19.08.2010
	 */
	template <typename T>
	class EventType
	{
	public:
		/**
		 * Name of the EventType. This name can take any value and is just
		 * for ease-of-debug for us, poor humans.
		 **/
		const string name;

		/**
		 * Creates a new EventType instance with the specified name.
		 *
		 * \param name	The name of the EventType
		 */
		EventType(const string& name) : name(name) {}

		/**
		 * Creates a new EventType instance which is a copy of the specified EventType instance.
		 *
		 * \param evt	The EventType to match.
		 */
		EventType(const EventType<T>& evt) : name(evt.name) {}

		/**
		 * Simply destroys this EventType.
		 */
		virtual ~EventType(void) {}
	};

	/**
	 * \brief Map between a listener and the EventType it listen to (internal).
	 *
	 * This is an internal class used to store and sort the listeners when
	 * needed in an EventDispatcher.
	 *
	 * \author	Cedric Tabin
	 * \date	20.08.2010
	 */
	template <typename T>
	class EventLinker
	{
	public:
		/**
		 * Listener associated to the EventType. 
		 */
		const T listener;

		/**
		 * EventType referenced by the listener.
		 */
		const EventType<T>* eventType;

		/**
		 * Priority of the listener in the queue.
		 */
		int priority;

		/**
		 * Unique index for the listener. This value
		 * is used to keep an the insert-order in the listeners
		 * of same priority.
		 */
		unsigned int index;

		/**
		 * Creates a new EventLinker instance.
		 *
		 * \param eventType	The EventType.
		 * \param listener	The listener.
		 * \param priority	The priority.
		 * \param index		The unique index.
		 */
		EventLinker(const EventType<T>* eventType, const T& listener, int priority, unsigned int index) : listener(listener), eventType(eventType), priority(priority), index(index) {}

		/**
		 * Creates a new EventLinker instance which is a copy of the specified EventLinker instance.
		 *
		 * \param evt	The EventLinker to copy.
		 */
		EventLinker(const EventLinker<T>& evt) : listener(evt.listener), eventType(evt.eventType), priority(evt.priority), index(evt.index) {}

		/**
		 * Destroys this EventLinker. This will not free/delete the listener and the
		 * EventType it references.
		 */
		virtual ~EventLinker(void) {}

		/**
		 * Returns true if this EventLinker instance is smaller (or before than) the specified
		 * instance. Note that a smaller instance means a higher value for the priority field
		 * or, if the priority fields are equal, a lower value for the index field.
		 *
		 * \param evt	The EventLinker instance to compare to.
		 */
		bool operator< (const EventLinker& evt) const;
	};
    
    //------------------------------------------------------------------
    static bool eventlinker_comparator(EventLinker<void*>* a, EventLinker<void*>* b)
    {
        return (*a) < (*b);
    }

    template<typename T>
    bool EventLinker<T>::operator<(const EventLinker& evt) const
    {
        if (evt.priority < priority) return true;
        if (evt.priority > priority) return false;
        return index < evt.index;
    }
    //------------------------------------------------------------------

	/**
	 * \brief Root abstract class for all events.
	 *
	 * An AbstractEvent instance can be dispatched through an EventDispatcher. All the sub-classes
	 * of AbstractEvent must implement the notifyListener method that will call the listener's right
	 * method.
	 *
	 * \author	Cedric Tabin
	 * \date	24.08.2010
	 */
	template <typename T>
	class AbstractEvent
	{
	friend class EventDispatcher;
	public:
		/**
         * Defines the EventType of the AbstractEvent.
		 * \return The event type.
		 */
		const EventType<T>* getType(void) const { return type; }

		/**
         * Returns the EventDispatcher who is dispatching the AbstractEvent. If
         * the AbstractEvent is currently not begin dispatched, this value is NULL.
         * This value will hold the source EventDispatcher after the dispatching has
         * finished.
		 * \return The EventDispatcher who is dispatching the AbstractEvent.
		 */
		const EventDispatcher* getSource(void) const { return source; }

		/**
		 * Tells the AbstractEvent to stop immediatly its propagation. It is then
		 * the AbstractEvent::notifyListener method implementation to stop the EventDispatcher by
		 * returning false.
		 */
		void stopPropagation(void) { propagate = false; }

		/**
		 * Returns true if the AbstractEvent::stopPropagation method has been called.
		 *
		 * \return true if the propagation of the Event must be stopped.
		 */
		bool isPropagationStopped(void) const { return !propagate; }

		/**
		 * Creates a new AbstractEvent instance.
		 *
		 * \param eventType	The type of event.
		 */
		AbstractEvent(const EventType<T>* eventType) : type(eventType), source(NULL), propagate(true) {}

		/**
		 * Creates a new AbstractEvent instance which is a copy of the specified AbstractEvent instance.
		 *
		 * \param evt	The Event to copy.
		 */
		AbstractEvent(const AbstractEvent<T>& evt) : type(evt.type), source(evt.source), propagate(evt.propagate) {}

		/**
		 * Destroys this AbstractEvent. This will not free/delete the eventType.
		 */
		virtual ~AbstractEvent(void) {}

	protected:
		/**
		 * Process the listener notification.
		 *
		 * \param listener  The listener to invoke.
         * \param priority  The listener priority.
		 * \return true if the dispatching of the Event must continue. If false is returned,
		 * 	the event dispatching will be stopped immediately.
		 */
		virtual bool notifyListener(T listener, int priority) = 0;

	private:
        const EventType<T>* type;
        const EventDispatcher* source;
		bool propagate;
	};

	/**
	 * \brief Manages the dispaching of an event through listeners.
	 *
	 * \author	Cedric Tabin
	 * \date	21.08.2010
	 */
	class EventDispatcher
	{
	public:
		/**
		 * Add a listener to the specified EventType. The priority can take any integer values.
		 * There also three constants defined : HIGH_PRIORITY (100), MEDIUM_PRIORITY (50), NORMAL_PRIORITY (0) 
		 * and LOW_PRIORITY (-100).
		 *
		 * The listener S must be a subtype of T, otherwise there will be a compiler error.
		 *
		 * \param eventType	The EventType to listen to.
		 * \param listener	The listener to be invoked.
		 * \param priority	The priority of the listener.
		 */
		template <typename T, typename S>	
		void addListener(const EventType<T>* eventType, const S& listener, int priority=NORMAL_PRIORITY);

		/**
		 * Removes a listener from the specified EventType. If the listener appears
		 * many times, only the first instance will be removed. This method won't
		 * free or delete the EventType nor the listener.
		 *
		 * The listener S must be a subtype of T, otherwise there will be a compiler error.
		 *
		 * \param eventType	The EventType to remove the listener from.
		 * \param listener	The listener to remove.
		 * \return true if a listener has been removed.
		 */
		template<typename T, typename S>
		bool removeListener(const EventType<T>* eventType, const S& listener);

		/**
		 * Dispatches an event through the listeners.
		 *
		 * \param event		The AbstractEvent to dispatch.
		 * \return true if the event has been successfully dispatched, false if
		 * 	its propagation has been stopped (AbstractEvent::stopPropagation). If there is 
         *  no listener, true will be returned.
		 */
		template<typename T>
		bool dispatchEvent(AbstractEvent<T>* event) const;

		/**
		 * Returns true if some listeners will be notified for the specified
		 * event.
		 *
		 * \param eventType	The EventType to check.
		 * \return true if there exist one or more listener for the event.
		 */
		template<typename T>
		bool willTrigger(const EventType<T>* eventType) const;

		/**
		 * Removes all the listeners attached to the specified EventType. It is the developer's
		 * responsability to delete/free the listeners and EventType instances.
		 *
		 * \param eventType	The EventType to remove the listeners from.
		 */
		template<typename T>
		void clear(const EventType<T>* eventType);

		/**
		 * Removes all the listeners from this EventDispatcher. It is the developer's responsability to
		 * delete/free the listeners and EventType instances.
		 */
		void clear(void);

		/**
		 * Creates a new EventDispatcher instance.
		 */
		EventDispatcher(void) {}

		/**
		 * Creates a copy of the specified EventDispatcher. The new EventDispatcher instance
		 * will contains all a direct reference to the listeners of the source EventDispatcher. Thus,
		 * by modifying this instance, the copied instance will also be modified.
		 *
		 * \param e	The EventDispatcher instance to copy.
		 */
		EventDispatcher(const EventDispatcher& e) : listeners(e.listeners) {};

		/**
		 * Destroys this EventDispatcher instance. It is the developer's responsability
		 * to delete/free the listeners and EventType instances.
		 */
		virtual ~EventDispatcher(void) { clear(); }

	private:
		map<void*, list<EventLinker<void*>*> > listeners;
	};
    
    //------------------------------------------------------------------
    template<typename T, typename S>
    void EventDispatcher::addListener(const EventType<T>* eventType, const S& listener, int priority)
    {
        static int static_index = 0;

        // Relationship S -> T is done by this line at compile-time
        EventLinker<T>* evtLinker = new EventLinker<T>(eventType, listener, priority, static_index);
        
        list<EventLinker<void*>*>& evtListeners = listeners[(void*)eventType];
        evtListeners.push_back((EventLinker<void*>*)evtLinker);
        evtListeners.sort(eventlinker_comparator);
        
        static_index++;
    }

    template<typename T, typename S>
    bool EventDispatcher::removeListener(const EventType<T>* eventType, const S& listener)
    {
        if (!willTrigger(eventType)) return false;

        list<EventLinker<void*>*>& evtListeners = listeners[(void*)eventType];
        typename list<EventLinker<void*>*>::iterator it;
        for (it=evtListeners.begin() ; it!=evtListeners.end() ; it++)
        {
            EventLinker<T>* evtLink = (EventLinker<T>*) *it;

            if (evtLink->listener == listener)
            {
                evtListeners.erase(it);
                delete evtLink;

                //if there is no more listeners for the event type, remove it from
                //the list
                if (evtListeners.size() == 0) listeners.erase((void*)eventType);

                return true;
            }
        }

        return false;
    }

    template <typename T>
    bool EventDispatcher::willTrigger(const EventType<T>* eventType) const
    {
        return listeners.find((void*)eventType) != listeners.end();
    }

    template <typename T>
    bool EventDispatcher::dispatchEvent(AbstractEvent<T>* event) const
    {
        event->source = this;

        if (!willTrigger(event->type)) return true;

        list<EventLinker<void*>*> evtListeners = listeners.find((void*)(event->type))->second;
        typename list<EventLinker<void*>*>::const_iterator it;
        for (it=evtListeners.begin() ; it!=evtListeners.end() ; it++)
        {
            EventLinker<T>* evtLinker = (EventLinker<T>*) *it;
            bool continuePropagation = event->notifyListener(evtLinker->listener, evtLinker->priority);
            if (!continuePropagation) return false;
        }

        return true;
    }

    template <typename T>
    void EventDispatcher::clear(const EventType<T>* eventType)
    {
        list<EventLinker<void*>*>& evtListeners = listeners[(void*)eventType];

        list<EventLinker<void*>*>::iterator itl;
        for (itl=evtListeners.begin() ; itl!=evtListeners.end() ; itl++)
        {
            EventLinker<void*>* evtLinker = (*itl);
            delete evtLinker;
        }

        evtListeners.clear();
        listeners.erase((void*)eventType);
    }

    void EventDispatcher::clear(void)
    {
        map<void*, list<EventLinker<void*>*> >::iterator itm;
        for (itm=listeners.begin() ; itm!=listeners.end() ; itm++)
        {
            void* eventType = itm->first;
            clear((EventType<void*>*)eventType);
        }

        listeners.clear();
    }
    //------------------------------------------------------------------

	/**
	 * \brief Listener that can be notified of an Event.
	 *
	 * \author	Cedric Tabin
	 * \date	21.08.2010
	 */
	class EventListener
	{
	public:
		/**
		 * Method that will be invoked when an Event is dispatched through an EventDispatcher.
		 *
		 * \param event	Event performed.
		 */
		virtual void eventPerformed(Event* event) = 0;
	};

	/**
	 * \brief Global event class.
	 *
	 * An Event can be used for many types of notification.
	 *
	 * \author	Cedric Tabin
	 * \date	24.08.2010
	 */
	class Event : public AbstractEvent<EventListener*>
	{
	public:

		/**
		 * Notification event.
		 */
		static const EventType<EventListener*>* NOTIFY;
        
        /**
         * Open event.
         */
        static const EventType<EventListener*>* OPEN;
        
        /**
         * Close event.
         */
        static const EventType<EventListener*>* CLOSE;
        
        /**
         * Init event.
         */
        static const EventType<EventListener*>* INIT;
        
        /**
         * Complete event.
         */
        static const EventType<EventListener*>* COMPLETE;
        
		/**
		 * Creates a new Event instance.
		 *
		 * \param eventType	The event type
		 */
		Event(const EventType<EventListener*>* eventType) : AbstractEvent<EventListener*>(eventType) {}

		/**
		 * Creates a new Event instance which is a copy of the specified Event instance.
		 *
		 * \param evt	The instance to copy.
		 */
		Event(const Event& evt) : AbstractEvent<EventListener*>(evt) {}

		/**
		 * Destroys the current Event instance.
		 */
		virtual ~Event(void) {}

	protected:
		/**
		 * Calls EventListener::eventPerformed method on the specified listener.
		 *
		 * \param listener	The listener to invoke.
         * \param priority  The priority of the listener.
		 * \return true if the event propagation hasn't been stopped (AbstractEvent::stopPropagation),
         * meaning that the propagation must continue.
		 */
		virtual bool notifyListener(EventListener* listener, int priority);
	};
    
    //------------------------------------------------------------------
    const EventType<EventListener*>* Event::NOTIFY = new EventType<EventListener*>("notify");
    const EventType<EventListener*>* Event::OPEN = new EventType<EventListener*>("open");
    const EventType<EventListener*>* Event::CLOSE = new EventType<EventListener*>("close");
    const EventType<EventListener*>* Event::INIT = new EventType<EventListener*>("init");
    const EventType<EventListener*>* Event::COMPLETE = new EventType<EventListener*>("complete");

    bool Event::notifyListener(EventListener* listener, int priority)
    {
        listener->eventPerformed(this);
        return !isPropagationStopped();
    }
    //------------------------------------------------------------------
}

#endif

